import doctest
import itertools

try:
    from unittest import mock
except ImportError:  # python 3.2 or lower
    import mock

from django.test import TestCase
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from django.conf import settings

from tictactoe.contest import logic
from tictactoe.contest.templatetags import contest_extras
from tictactoe.tools.testing import sync_celery

from .models import Entry, Fight, LatestEntry
from . import views


class EntryTestCase(TestCase):
    setUp = lambda self: new_user(self)

    def test_code_constraint_over(self):
        """code less than allowed chars, more than allowed bytes"""
        # 'š' is a 2-byte utf8 character
        code = "".join(['š'] * (settings.MAX_CODE_SIZE // 2 + 1))
        e = Entry.objects.create(user=self.user1, code=code)
        self.assertRaises(ValidationError, e.full_clean)

    def test_code_constraint_just_enough(self):
        # 'š' is a 2-byte utf8 character
        code = "".join(['š'] * (settings.MAX_CODE_SIZE // 2))
        e = Entry.objects.create(user=self.user1, code=code)
        self.assertIsNone(e.full_clean())

    def test_code_size(self):
        code = 'žžž'
        e = Entry.objects.create(user=self.user1, code=code)
        self.assertEqual(6, e.codesize)


@sync_celery
class ViewTestCase(TestCase):
    setUp = lambda self: new_user(self)

    def test_index(self):
        request = self.client.get('/')
        self.assertEqual(request.status_code, 200)

    def test_upload_get(self):
        self.client.login(username='u1', password='u1')
        response = self.client.get(reverse(views.upload))
        self.assertEqual(response.status_code, 200)

    def test_upload_post(self):
        self.assertTrue(self.client.login(username='u1', password='u1'))
        self.assertEqual(1, Entry.objects.count())
        self.client.post(reverse(views.upload), {'code': 'lua1'})
        self.assertEqual(2, Entry.objects.count())


class ResultsTestCase(TestCase):
    def setUp(self):
        new_user(self)
        new_entry(self)
        self.l1 = LatestEntry.objects.create(user=self.user1, entry=self.e1)
        self.l2 = LatestEntry.objects.create(user=self.user2, entry=self.e2)

    def test_draw(self):
        Fight(x=self.e1, o=self.e2, result='draw').save()
        Fight(x=self.e2, o=self.e1, result='draw').save()
        self.assertEqual((0, 2, 0), self.l1.calc_results())
        self.assertEqual((0, 2, 0), self.l2.calc_results())

    def test_win_loss(self):
        Fight(x=self.e1, o=self.e2, result='win').save()
        self.assertEqual((1, 0, 0), self.l1.calc_results())
        self.assertEqual((0, 0, 1), self.l2.calc_results())


@sync_celery
class CeleryFightTestCase(TestCase):
    patch_x_wins = itertools.cycle((('ok', 'x', []), ('ok', 'o', [])))

    setUp = lambda self: new_user(self) or new_entry(self)

    def test_qualify(self):
        self.e1.qualify()
        self.l1 = LatestEntry.objects.create(user=self.user1, entry=self.e1)
        self.assertEqual((0, 0, 0), self.l1.calc_results())

    @mock.patch('tictactoe.contest.tasks.compete', side_effect=patch_x_wins)
    def test_two_users_draw(self, patch):
        self.e1.qualify()
        l1 = LatestEntry.objects.get(entry=self.e1)
        self.assertEqual((0, 0, 0), l1.calc_results())

        self.e2.qualify()
        # e2 qualified later, so won against e1
        l2 = LatestEntry.objects.get(entry=self.e2)
        self.assertEqual((2, 0, 0), l2.calc_results())
        # e1 qualified earlier, so lost against e2
        self.assertEqual((0, 0, 2), l1.calc_results())
        # bot never has latestentry
        self.assertEqual(2, LatestEntry.objects.count())

        # requalification preserves the calc_results()
        self.e2.qualify()
        self.assertEqual((2, 0, 0), l2.calc_results())
        self.assertEqual((0, 0, 2), l1.calc_results())


# ============================================================================
# Helpers
# ============================================================================


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(logic))
    setUp = lambda self: new_user(self) or new_entry(self)
    tests.addTests(doctest.DocTestSuite(
        contest_extras, setUp=setUp))
    return tests


def new_user(self):
    self.user1 = User.objects.create(username='u1')
    self.user1.set_password('u1')
    self.user1.save()
    self.user2 = User.objects.create(username='u2')
    self.user2.set_password('u2')
    self.user2.save()
    if hasattr(self, 'globs'):
        self.globs.update({'user1': self.user1, 'user2': self.user2})


def new_entry(self):
    code1 = 'lua code'
    self.e1 = Entry.objects.create(user=self.user1, code=code1)
    self.e2 = Entry.objects.create(user=self.user2, code=code1)
    if hasattr(self, 'globs'):
        self.globs.update({'e1': self.e1, 'e2': self.e2})



# encoding: utf8
from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contest', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Fight',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, verbose_name='ID', auto_created=True)),
                ('x', models.ForeignKey(to_field='id', to='contest.Entry')),
                ('o', models.ForeignKey(to_field='id', to='contest.Entry')),
                ('gameplay', models.CommaSeparatedIntegerField(max_length=230, help_text="Gameplay flow. Board is separated to 81 cells. Each number means a move by alternating player. For example, '10,1,0' means: x placed (2,1,1,1), o placed (1,1,1,1) and x made an error. In case 0 is at the end (like in the example), 'error' field is non-empty.")),
                ('error', models.CharField(max_length=255, help_text="Non-empty if `gameplay' ends with zero", blank=True)),
                ('result', models.CharField(max_length=10, help_text='Fight result of x (e1) versus o (e2). Relative to e1.', choices=[('win', 'win'), ('draw', 'draw'), ('loss', 'loss')])),
            ],
            options={
                'unique_together': set([('x', 'o')]),
                'index_together': set([('x', 'result'), ('o', 'result')]),
            },
            bases=(models.Model,),
        ),
    ]



from django.utils.translation import ugettext_lazy as _
from django.template import RequestContext
from django.shortcuts import redirect, render_to_response, get_object_or_404
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.db.models import Q

from .forms import CodeUploadForm
from .models import Entry, LatestEntry, Fight, HandedOutCaps


def who_got_capped(request):
    caps = HandedOutCaps.objects.all()
    return render_to_response(
        'contest/who_got_capped.html', {'caps': caps},
        context_instance=RequestContext(request))


def entry(request, id):
    entry = get_object_or_404(Entry, pk=id)
    fights = Fight.objects.filter(Q(x=entry) | Q(o=entry)).order_by('id').all()
    return render_to_response(
        'contest/entry.html', {'entry': entry, 'fights': fights},
        context_instance=RequestContext(request))


def entries(request, uid=None):
    if uid is None:
        entries = Entry.objects
        user = None
    else:
        entries = Entry.objects.filter(user__id=uid)
        user = User.objects.get(id=uid)
    return render_to_response(
        'contest/entries.html', {'entries': entries.all(), 'user': user},
        context_instance=RequestContext(request))


def fight(request, id):
    fight = get_object_or_404(Fight, pk=id)
    return render_to_response(
        'contest/fight.html', {'fight': fight},
        context_instance=RequestContext(request))


def ranking(request):
    # Piggy-back 'results' to every value
    latest = LatestEntry.objects.all()
    res = [e.calc_results() for e in latest]
    [setattr(l, 'results', r) for l, r in zip(latest, res)]
    latest2 = sorted(latest, key=lambda x: x.results[0], reverse=True)

    return render_to_response(
        'contest/ranking.html', {'latestentries': latest2},
        context_instance=RequestContext(request))


@login_required
def upload(request):
    if request.method == 'POST':
        form = CodeUploadForm(request.POST)
        if form.is_valid():
            entry = form.instance
            entry.user = request.user
            entry.save()
            entry.qualify()
            messages.success(request, _("Code uploaded"))
            return redirect(entry.get_absolute_url())
    else:
        form = CodeUploadForm()
    return render_to_response(
        'contest/upload.html', {'form': form},
        context_instance=RequestContext(request))



from django import template

register = template.Library()


@register.filter
def result_of(fight, user):
    """Returns result of a fight relative to the user.

    So this:

        User {{ entry.user }}
            {% if fight.x.user == entry.user %}
                {{ fight.result_of_x }}
            {% else %}
                {{ fight.result_of_o }}
            {% endif %}
        the match!

    Can be replaced by:

        User {{ entry.user }} {{ fight|result_of:entry.user }} the match!

    >>> from tictactoe.contest.models import Fight
    >>> result_of(Fight(x=e1, o=e2, result='win'), user1)
    'win'
    >>> result_of(Fight(x=e1, o=e2, result='win'), user2)
    'loss'
    >>> result_of(Fight(x=e1, o=e2, result='loss'), user1)
    'loss'
    >>> result_of(Fight(x=e1, o=e2, result='loss'), user2)
    'win'
    >>> result_of(Fight(x=e1, o=e2, result='draw'), user1)
    'draw'
    >>> result_of(Fight(x=e1, o=e2, result='draw'), user2)
    'draw'
    """
    if fight.x.user == user:
        return fight.result
    else:
        if fight.result == 'draw':
            return 'draw'
        return fight.result == 'loss' and 'win' or 'loss'



from django.views.generic import TemplateView
from django.conf.urls import patterns, include, url
from django.contrib import admin


urlpatterns = patterns(
    '',
    url(r'^$', TemplateView.as_view(template_name='contest/index.html'),
        name='index'),
    url(r'rules$', TemplateView.as_view(template_name='contest/rules.html'),
        name='rules'),

    url(r'^who_got_capped$', 'tictactoe.contest.views.who_got_capped'),
    url(r'^upload$', 'tictactoe.contest.views.upload'),
    url(r'^ranking$', 'tictactoe.contest.views.ranking'),
    url(r'^entries$', 'tictactoe.contest.views.entries'),
    url(r'^entries/(?P<uid>\d+)$', 'tictactoe.contest.views.entries'),
    url(r'^entry/(?P<id>\d+)$', 'tictactoe.contest.views.entry'),
    url(r'^fight/(?P<id>\d+)$', 'tictactoe.contest.views.fight'),

    url(r'^login/$', 'django.contrib.auth.views.login', name='login'),
    url(r'^logout$', 'django.contrib.auth.views.logout',
        {'next_page': '/'}, name='logout'),
    url(r'^admin/', include(admin.site.urls)),

)



import logging

from django.core.serializers import serialize, deserialize
from django.conf import settings

from tictactoelib import compete

from tictactoe.celery_app import app

from .models import Fight, Entry, LatestEntry
from .logic import winner


logger = logging.getLogger(__name__)


def fight(e1, e2):
    logmsg = "Entry %d (%s) vs entry %d (%s) complete: %s, %s"
    e1log = e1.id, e1.user
    e2log = e2.id, e2.user

    kwargs = dict(
        cgroup=settings.FIGHT_CGROUP,
        memlimit=settings.FIGHT_MEMORY_LIMIT,
        timeout=settings.FIGHT_TIMEOUT,
    )
    round1 = compete(e1.code, e2.code, **kwargs)
    logger.debug(logmsg % (e1log + e2log + tuple(round1[:2])))

    round2 = compete(e2.code, e1.code, **kwargs)
    logger.debug(logmsg % (e2log + e1log + tuple(round2[:2])))

    Fight.from_compete(e1, e2, round1).save()
    Fight.from_compete(e2, e1, round2).save()
    return round1, round2


@app.task
def schedule_qualification(e1_json):
    e1 = next(deserialize('json', e1_json)).object
    args1 = e1.id, e1.user
    logger.debug("Executing qualification for entry %d (%s)" % args1)
    dumb = Entry.qualification_entry()
    round1, round2 = fight(e1, dumb)

    if winner(round1, round2) == 'e1':
        e1.add_latest()  # this entry is able to compete now
        # Schedule pvp
        les = LatestEntry.objects.exclude(user=e1.user).exclude(user=dumb)
        for latestentry in les.select_related('entry').all():
            e2 = latestentry.entry
            args2 = e1.id, e1.user, e2.id, e2.user
            logger.debug(("Scheduling competition between entries %d (%s) "
                          "and %d (%s)") % args2)
            e1_json = serialize('json', [e1])
            e2_json = serialize('json', [e2])
            schedule_compete.delay(e1_json, e2_json)
    else:
        logger.info("Entry %d (%s) did not qualify" % (e1.id, e1.user))


@app.task
def schedule_compete(e1_json, e2_json):
    e1 = next(deserialize('json', e1_json)).object
    e2 = next(deserialize('json', e2_json)).object
    fight(e1, e2)



import os

from celery import Celery

from django.conf import settings


os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'tictactoe.settings')
app = Celery('tictactoe')
app.config_from_object('django.conf:settings')
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)



import os
from tictactoe.config import config
from django.core.urlresolvers import reverse_lazy

APP_DIR = os.path.dirname(__file__)
BASE_DIR = os.path.dirname(APP_DIR)

DEBUG = config.getboolean('tictactoe', 'debug')

TEMPLATE_DEBUG = DEBUG

SECRET_KEY = config.get('tictactoe', 'secret_key', fallback='supersecret')

ALLOWED_HOSTS = ['tictactoe.spilgames.com']

ADMINS = (('Motiejus', 'motiejus.jakstys@gmail.com'),)
MANAGERS = ADMINS

DEV_APPS = (
    'django_extensions',
    'debug_toolbar'
)

INSTALLED_APPS = (
    'django.contrib.contenttypes',
    'django.contrib.auth',
    'django.contrib.admin',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    'jquery',
    'bootstrap',

    'tictactoe.tools',
    'tictactoe.contest',
) + (DEBUG and DEV_APPS or ())

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

TEMPLATE_CONTEXT_PROCESSORS = (
    'django.contrib.auth.context_processors.auth',
    'django.core.context_processors.debug',
    'django.core.context_processors.i18n',
    'django.core.context_processors.media',
    'django.core.context_processors.static',
    'django.core.context_processors.tz',
    'django.contrib.messages.context_processors.messages',
    'django.core.context_processors.request',
    'tictactoe.tools.context_processors.debug',
)

TEMPLATE_DIRS = (
    os.path.join(APP_DIR, 'templates'),
)

ROOT_URLCONF = 'tictactoe.urls'

WSGI_APPLICATION = 'tictactoe.wsgi.application'

_DBNAME, _DBENGINE = config.get('db', 'name'), config.get('db', 'engine')
if DEBUG and "sqlite" in _DBENGINE and "/" not in _DBNAME:
    _DBNAME = os.path.join(BASE_DIR, _DBNAME)

DATABASES = {
    'default': {
        'ENGINE': _DBENGINE,
        'NAME': _DBNAME,
        'USER': config.get('db', 'user'),
        'PASSWORD': config.get('db', 'password'),
        'HOST': config.get('db', 'host'),
    }
}

# Getting international

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files

STATIC_URL = '/static/'

STATICFILES_DIRS = (os.path.join(APP_DIR, 'static'),)

STATIC_ROOT = config.get('tictactoe', 'static_root') or \
    os.path.join(BASE_DIR, 'static')

LOGIN_URL = reverse_lazy('login')
LOGIN_REDIRECT_URL = '/'

LOGGING = {
    'version': 1,
    'formatters': {
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
        'timed': {
            'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        },
    },
    'handlers': {
        'null': {
            'level': 'DEBUG',
            'class': 'logging.NullHandler',
        },
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
        },
        'syslog': {
            'level': 'DEBUG',
            'class': 'logbook.compat.RedirectLoggingHandler',
            'formatter': 'simple',
        },
    },
    'loggers': {
        'tictactoe': {
            'handlers': ['console', 'syslog'],
            'propagate': True,
            'level': 'INFO',
            },
        'django': {
            'handlers': ['console', 'syslog'],
            'propagate': True,
            'level': 'INFO',
            }
        }
    }

# Maximum size of Lua code submission
MAX_CODE_SIZE = 60000

# Celery

CELERY_TASK_SERIALIZER = 'msgpack'
CELERY_ACCEPT_CONTENT = ['msgpack', 'yaml']
BROKER_URL = 'redis://'

FIGHT_CGROUP = config.get('limits', 'cgroup')
_FIGHT_MEMORY_LIMIT = config.get('limits', 'memory')
FIGHT_MEMORY_LIMIT = _FIGHT_MEMORY_LIMIT and int(_FIGHT_MEMORY_LIMIT) or None
FIGHT_TIMEOUT = config.getfloat('limits', 'timeout')


import logbook
import logbook.handlers

# add syslog handler
syslog_handler = logbook.handlers.SyslogHandler(
    application_name='tictactoe', facility='local6',
    address=('localhost', 514)
    )
syslog_handler.push_application()



from django.db import models


class OwnManager(models.Manager):
    def __init__(self, key):
        super(OwnManager, self).__init__()
        self._key = key

    def __call__(self, user, *args, **kwargs):
        self._user = user
        return self

    def get_queryset(self):
        assert hasattr(self, '_user'), "must call Manager with user instance"
        if self._user.is_superuser:
            return super(OwnManager, self).get_queryset()
        d = {self._key: self._user}
        return super(OwnManager, self).get_queryset().filter(**d)



import logging
import os

import configparser

config = configparser.RawConfigParser()

HERE = os.path.dirname(__file__)

configfiles = [
    os.path.join(HERE, 'tictactoe.default.cfg'),  # default config
    '/etc/tictactoe.cfg',                         # staging/live
    os.path.join(HERE, 'tictactoe.cfg'),          # per-environment config
]

files_read = config.read(configfiles)

logging.debug('config files read: %s' % files_read)



# encoding: utf8
from django.db import models, migrations
from django.conf import settings
import tictactoe.tools.validators


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Entry',
            fields=[
                ('id', models.AutoField(primary_key=True, verbose_name='ID', serialize=False, auto_created=True)),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL, to_field='id')),
                ('code', models.TextField(validators=[tictactoe.tools.validators.ByteLengthValidator(60000)])),
                ('uploaded', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



import logging

from django.db import models
from django.core.urlresolvers import reverse_lazy
from django.core.serializers import serialize
from django.contrib.auth.models import User
from django.utils.translation import ugettext_lazy as _
from django.template.defaultfilters import date, filesizeformat as fmt
from django.conf import settings

from tictactoe.tools.validators import ByteLengthValidator
from tictactoe.tools.models import OwnManager

from . import fixtures


logger = logging.getLogger(__name__)


class HandedOutCaps(models.Model):
    user = models.ForeignKey(User)
    when = models.DateTimeField(auto_now_add=True)
    notes = models.CharField(max_length=255)

    def __str__(self):
        return "To %s at %s" % (self.user, date(self.when, "Y-m-d H:i:s"))

    class Meta:
        ordering = '-when',


class Entry(models.Model):
    _max_len = ByteLengthValidator(settings.MAX_CODE_SIZE)

    user = models.ForeignKey(User)
    code = models.TextField(validators=[_max_len])

    uploaded = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    objects = models.Manager()
    own = OwnManager('user')

    @property
    def codesize(self):
        return len(self.code.encode('utf8'))

    def get_absolute_url(self):
        return reverse_lazy('tictactoe.contest.views.entry', args=[self.id])

    def __str__(self):
        return "<Entry %d by %s (%s)>" % (
            self.id, self.user, fmt(self.codesize))

    def add_latest(self):
        """After qualifying, make it LatestEntry of this user"""
        latest = LatestEntry.objects.filter(user=self.user).first()
        if latest is None:
            LatestEntry(user=self.user, entry=self).save()
        else:
            latest.entry = self
            latest.save()

    def qualify(self):
        """Scheulde a qualification with dumb_player and maybe go compete"""
        from .tasks import schedule_qualification
        logger.info("Scheduling %s for qualification" % self)
        schedule_qualification.delay(serialize('json', [self]))

    @staticmethod
    def qualification_entry():
        u = User.objects.get(username='Qualification-Bot')
        return Entry.objects.get(user=u)

    class Meta:
        verbose_name_plural = 'Entries'
        ordering = '-uploaded',


class LatestEntry(models.Model):
    user = models.ForeignKey(User)
    entry = models.ForeignKey(Entry)

    def calc_results(self):
        r1 = Fight.objects.filter(
            x__latestentry=self, o__latestentry__isnull=False
            ).values('result').annotate(cnt=models.Count('result'))
        r2 = Fight.objects.filter(
            o__latestentry=self, x__latestentry__isnull=False
            ).values('result').annotate(cnt=models.Count('result'))
        d1 = dict(((d['result'], d['cnt']) for d in r1))
        d2 = dict(((d['result'], d['cnt']) for d in r2))
        win = d1.get('win', 0) + d2.get('loss', 0)
        draw = d1.get('draw', 0) + d2.get('draw', 0)
        loss = d1.get('loss', 0) + d2.get('win', 0)
        return win, draw, loss

    def __str__(self):
        return "<LatestEntry by %s>" % (self.user)

    class Meta:
        verbose_name_plural = 'Latest Entries'


class Fight(models.Model):
    """e1 is x, e2 is o. Results relative to e1.

    Each Entry has two Fight objects: one with x, one with o"""

    FIGHT_RESULT = (
        ('win', _('win')),
        ('draw', _('draw')),
        ('loss', _('loss')),
    )

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    x = models.ForeignKey(Entry, related_name='e1')
    o = models.ForeignKey(Entry, related_name='e2')

    # len(",".join(map(str, range(1, 81)))) == 230
    # Zero signifies error and in case of error is always the last number
    gameplay = models.CommaSeparatedIntegerField(
        max_length=230,
        help_text=_(
            "Gameplay flow. Board is separated to 81 cells. Each "
            "number means a move by alternating player. For example, "
            "'10,1,0' means: x placed (2,1,1,1), o placed (1,1,1,1) and "
            "x made an error. In case 0 is at the end (like in the example), "
            "'error' field is non-empty.")
    )

    error = models.CharField(
        max_length=255, blank=True,
        help_text=_("Non-empty if `gameplay' ends with zero")
    )

    result = models.CharField(
        max_length=10, choices=FIGHT_RESULT,
        help_text=_("Fight result of x (e1) versus o (e2). Relative to e1.")
    )

    def result_of_x(self):
        return self.result

    def result_of_o(self):
        if self.result == 'draw':
            return 'draw'
        return self.result == 'loss' and 'win' or 'win'

    class Meta:
        index_together = (
            ('x', 'result'),
            ('o', 'result'),
        )
        unique_together = ('x', 'o')

    @staticmethod
    def from_compete(x, o, round_result):
        """Creates Fight instance from output of tictactoelib.compete"""
        fight, _created = Fight.objects.get_or_create(x=x, o=o)
        if round_result[0] == 'ok':
            _, xodraw, gameplay = round_result
            res = 'draw'
            if xodraw == 'x' or xodraw == 'o':
                res = xodraw == 'x' and 'win' or 'loss'
            fight.gameplay = gameplay
            fight.result = res
            return fight
        elif round_result[0] == 'error':
            _, xo, reason, gameplay = round_result
            res = 'win' if xo == 'o' else 'loss'
            fight.gameplay = gameplay
            fight.error = reason
            fight.result = res
            return fight

    def get_absolute_url(self):
        return reverse_lazy('tictactoe.contest.views.fight', args=[self.id])

    def __str__(self):
        return "<Fight %s vs %s. Result: %s>" % (self.x, self.o, self.result)


models.signals.post_migrate.connect(fixtures.qualification_bot)



# encoding: utf8
from django.db import models, migrations
import datetime


class Migration(migrations.Migration):

    dependencies = [
        ('contest', '0003_latestentry'),
    ]

    operations = [
        migrations.AddField(
            model_name='fight',
            name='modified',
            field=models.DateTimeField(default=datetime.datetime(2014, 4, 7, 18, 8, 23, 959301), auto_now=True),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='fight',
            name='created',
            field=models.DateTimeField(default=datetime.datetime(2014, 4, 7, 18, 8, 28, 908912), auto_now_add=True),
            preserve_default=False,
        ),
    ]



# encoding: utf8
from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contest', '0004_auto_20140407_1808'),
    ]

    operations = [
        migrations.CreateModel(
            name='HandedOutCaps',
            fields=[
                ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True, serialize=False)),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL, to_field='id')),
                ('when', models.DateTimeField(auto_now_add=True)),
                ('notes', models.CharField(max_length=255)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



from django.contrib import admin

from .models import Entry, LatestEntry, Fight, HandedOutCaps


class LatestEntryAdmin(admin.ModelAdmin):
    list_display = ('user', 'entry')


class EntryAdmin(admin.ModelAdmin):
    actions = ['force_entry_qualification']
    list_display = ('user', 'codesize', 'uploaded')

    def force_entry_qualification(self, request, queryset):
        for e in queryset:
            e.qualify()
            self.message_user(request, "qualification rescheduled for %s." % e)


class FightAdmin(admin.ModelAdmin):
    list_display = ('x', 'o', 'result', 'error')
    list_filter = ('x__user', 'o__user')


admin.site.register(Entry, EntryAdmin)
admin.site.register(LatestEntry, LatestEntryAdmin)
admin.site.register(Fight, FightAdmin)
admin.site.register(HandedOutCaps)



# encoding: utf8
from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contest', '0002_fight'),
    ]

    operations = [
        migrations.CreateModel(
            name='LatestEntry',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, verbose_name='ID', auto_created=True)),
                ('user', models.ForeignKey(to_field='id', to=settings.AUTH_USER_MODEL)),
                ('entry', models.ForeignKey(to_field='id', to='contest.Entry')),
            ],
            options={
                'verbose_name_plural': 'Latest Entries',
            },
            bases=(models.Model,),
        ),
    ]



from django.test import override_settings


def sync_celery(function=None):
    """Synchronous task execution by celery. Does not require redis"""
    actual_decorator = override_settings(
        CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
        CELERY_ALWAYS_EAGER=True,
        BROKER_BACKEND='memory'
    )
    if function:
        return actual_decorator(function)
    return actual_decorator



from django.core.validators import BaseValidator
from django.utils.deconstruct import deconstructible
from django.utils.translation import ungettext_lazy


@deconstructible
class ByteLengthValidator(BaseValidator):
    compare = lambda self, a, b: a > b
    clean = lambda self, x: len(x.encode('utf8'))
    message = ungettext_lazy(
        ('Ensure this value has at most %(limit_value)d byte '
         '(it has %(show_value)d).'),
        ('Ensure this value has at most %(limit_value)d bytes '
         '(it has %(show_value)d).'),
        'limit_value')
    code = 'max_length'



"""Fight winner decision logic"""

__all__ = ['winner']


def flip(res):
    """flips 'x', 'o' or 'draw'."""
    if res == 'x':
        return 'o'
    elif res == 'o':
        return 'x'
    elif res == 'draw':
        return 'draw'
    else:
        raise RuntimeError("Invalid res: %s" % str(res))


def error_to_win(r):
    """Change one's error to other's win"""
    flip = lambda c: c == 'x' and 'o' or 'x'
    if r[0] == 'error':
        return flip(r[1])
    elif r[0] == 'ok':
        return r[1]
    else:
        raise RuntimeError("Invalid r[0]: %s", str(r[0]))


def winner(r1, r2):
    """Return winner of two rounds: 'e1' or 'e2' or 'draw'.

    round1 is e1 (x) vs e2 (o). round2 is e2(x) vs e1(o).

    Input format of r1 and r2:
    * 'ok', 'x' | 'draw' | 'o', ...
    * 'error', 'x' | 'o', ...

    >>> okx = 'ok', 'x'
    >>> oko = 'ok', 'o'
    >>> okdraw = 'ok', 'draw'
    >>> errx = 'error', 'x'
    >>> erro = 'error', 'o'

    >>> winner(okx, oko)
    'e1'
    >>> winner(okx, okx)
    'draw'
    >>> winner(oko, oko)
    'draw'
    >>> winner(okx, erro)
    'draw'
    >>> winner(okx, errx)
    'e1'
    >>> winner(errx, errx)
    'draw'
    >>> winner(erro, erro)
    'draw'
    >>> winner(errx, erro)
    'e2'
    >>> winner(okdraw, okdraw)
    'draw'
    >>> winner(okdraw, oko)
    'e1'
    >>> winner(okdraw, erro)
    'e2'
    """

    f1 = error_to_win(r1)
    f2 = error_to_win(r2)

    return {
        ('x', 'o'): 'e1',
        ('o', 'x'): 'e2',

        ('draw', 'o'): 'e1',
        ('o', 'draw'): 'e2',

        ('draw', 'x'): 'e2',
        ('x', 'draw'): 'e1',

        ('x', 'x'): 'draw',
        ('o', 'o'): 'draw',
        ('draw', 'draw'): 'draw',
        }[(f1, f2)]



import os
import sys

from setuptools import setup, find_packages

REQS_FILE = open(os.path.join(os.path.dirname(__file__), "requirements.txt"))
install_reqs = [r.strip() for r in REQS_FILE]

if sys.version_info < (3, 3):
    install_reqs.append("mock==1.0.1")

setup(
    name='tictactoe',
    version='0.1.4',
    author="Motiejus Jakštys",
    author_email="desired.mta@gmail.com",
    description="Ultimate Tic-Tac-Toe tictactoe front-end",
    long_description=open('README.rst').read(),
    url="https://github.com/Motiejus/tictactoe",
    packages=find_packages(),
    install_requires=install_reqs,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            "tictactoe = tictactoe.manage:main"
        ]
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Education",
        "Topic :: Games/Entertainment :: Puzzle Games",
        "License :: OSI Approved :: MIT License",
    ],
)



#!/usr/bin/env python
import csv

from django.core.management.base import BaseCommand, CommandError
from django.contrib.auth.models import User


class Command(BaseCommand):
    args = '<csv file>'
    help = 'Import users from dalyviai.csv'

    def handle(self, *args, **kwargs):
        try:
            csvfile = open(args[0], 'r')
        except IndexError:
            raise CommandError("Not enough arguments")
        except FileNotFoundError:
            raise CommandError("Cannot read '%s'" % args[0])
        reader = csv.DictReader(csvfile, delimiter=';')
        created = 0
        for row in reader:
            username, first_name = row['code'], row['name']
            surname, passwd = row['surname'], row['psw_day0']
            if User.objects.filter(username=username).exists():
                self.stdout.write("User %s exists, skipping" % username)
            else:
                User.objects.create_user(
                    username=username, password=passwd,
                    first_name=first_name, last_name=surname)
                self.stdout.write("User %s created" % username)
                created += 1
        self.stdout.write("%d user(s) created" % created)



from django.contrib.auth.models import User

from tictactoelib.examples import dumb_player


def qualification_bot(sender, **kwargs):
    from .models import Entry

    (q, created) = User.objects.get_or_create(username='Qualification-Bot')
    if Entry.objects.filter(user=q).first() is None:
        Entry.objects.create(user=q, code=dumb_player)
